pub mod math;
pub mod modules;
pub mod widget;

use std::sync::{
    Arc, Mutex,
    atomic::{AtomicU32, Ordering},
};

use gtk::{
    CssProvider, EventControllerKey, Orientation,
    gdk::{Display, Key},
    glib,
    prelude::*,
};
use gtk4_layer_shell::{Edge, Layer, LayerShell};
use nucleo_matcher::{Config, Matcher};

use crate::modules::{Module, SearchResult, desktop::DesktopModule};

fn main() -> glib::ExitCode {
    let app = gtk::Application::builder()
        .application_id("nl.kallestruik.application-launcher")
        .build();
    app.connect_startup(|_| load_css());
    app.connect_activate(build_ui);
    app.run()
}

fn build_ui(application: &gtk::Application) {
    let window = gtk::ApplicationWindow::new(application);
    window.set_title(Some("Application Launcher"));
    window.init_layer_shell();
    window.set_keyboard_mode(gtk4_layer_shell::KeyboardMode::Exclusive);
    window.set_layer(Layer::Overlay);

    // This centers the window
    window.set_anchor(Edge::Top, false);
    window.set_anchor(Edge::Right, false);
    window.set_anchor(Edge::Bottom, false);
    window.set_anchor(Edge::Left, false);

    let container = gtk::Box::builder()
        .css_classes(vec!["container"])
        .orientation(Orientation::Vertical)
        .height_request(600)
        .build();
    window.set_child(Some(&container));

    let input = gtk::Entry::builder()
        .css_classes(vec!["main-input"])
        .placeholder_text("Start typing...")
        .build();

    container.append(&input);

    let below_input_container = gtk::Box::builder()
        .orientation(Orientation::Horizontal)
        .spacing(0)
        .hexpand(true)
        .vexpand(true)
        .build();
    container.append(&below_input_container);

    let result_container = gtk::Box::builder()
        .css_classes(vec!["result-container"])
        .orientation(Orientation::Vertical)
        .spacing(0)
        .width_request(400)
        .overflow(gtk::Overflow::Hidden)
        .build();
    below_input_container.append(&result_container);

    let detail_container = gtk::Box::builder()
        .css_classes(vec!["detail-container"])
        .orientation(Orientation::Vertical)
        .spacing(2)
        .width_request(800)
        .build();
    below_input_container.append(&detail_container);

    let current_index = Arc::new(AtomicU32::new(0));
    let modules: Arc<Vec<Box<dyn Module>>> = Arc::new(vec![Box::new(DesktopModule::new())]);
    let matcher = Arc::new(Mutex::new(Matcher::new(Config::DEFAULT.match_paths())));
    let last_results = Arc::new(Mutex::new({
        let mut matcher = matcher.lock().unwrap();
        update_results("", &modules, &mut *matcher)
    }));

    input.connect_changed(glib::clone! {
        #[weak]
        result_container,
        #[weak]
        current_index,
        #[weak]
        last_results,
        move |input| {
            let search: &str = &input.text().to_lowercase();
            let mut matcher = matcher.lock().unwrap();
            current_index.store(0, Ordering::Relaxed);
            let mut last_results = last_results.lock().unwrap();
            *last_results = update_results(search, &modules, &mut *matcher);
            render_result_list(&result_container, last_results.iter().take(10), 0);
    }});

    input.connect_activate(glib::clone! {
        #[weak]
        current_index,
        #[weak]
        last_results,
        move|_| {
            let selected_index = current_index.load(Ordering::Relaxed);
            let last_results = last_results.lock().unwrap();
            match last_results.get(selected_index as usize) {
                Some(res) => res.execute(),
                None => {}
            }
            std::process::exit(0);
    }});

    let event_controller = EventControllerKey::new();
    {
        let last_results = last_results.clone();
        event_controller.connect_key_pressed(move |_, key, _, _| {
            match key {
                Key::Escape => {
                    std::process::exit(0);
                }
                Key::Down | Key::Tab => {
                    let last_results = last_results.lock().unwrap();
                    let _ = current_index.fetch_update(Ordering::SeqCst, Ordering::SeqCst, |i| {
                        if last_results.len() == 0 {
                            return Some(0);
                        }
                        if i as usize == (last_results.len() - 1).min(9) {
                            Some(i)
                        } else {
                            Some(i + 1)
                        }
                    });
                    let selected_index = current_index.load(Ordering::Relaxed);
                    render_result_list(
                        &result_container,
                        last_results.iter().take(10),
                        selected_index as usize,
                    );
                    return glib::Propagation::Stop;
                }
                Key::Up | Key::ISO_Left_Tab => {
                    let _ = current_index.fetch_update(Ordering::SeqCst, Ordering::SeqCst, |i| {
                        if i == 0 { Some(0) } else { Some(i - 1) }
                    });
                    let selected_index = current_index.load(Ordering::Relaxed);
                    let last_results = last_results.lock().unwrap();
                    render_result_list(
                        &result_container,
                        last_results.iter().take(10),
                        selected_index as usize,
                    );
                    return glib::Propagation::Stop;
                }
                _ => (),
            }
            glib::Propagation::Proceed
        });
    }

    window.add_controller(event_controller);

    window.show();
}

fn update_results(
    search: &str,
    modules: &[Box<dyn Module>],
    matcher: &mut Matcher,
) -> Vec<Box<dyn SearchResult>> {
    let mut results = vec![];
    {
        for module in modules.iter() {
            results.extend(module.results_for_search(search, matcher));
        }
    }
    results.sort_by(|a, b| b.confidence().cmp(&a.confidence()));
    results
}

fn render_result_list<'a>(
    container: &gtk::Box,
    results: impl Iterator<Item = &'a Box<dyn SearchResult>>,
    current_index: usize,
) {
    remove_all_children(&container);
    for (i, result) in results.enumerate() {
        widget::search_result::create(container, result, i == current_index);
    }
}

fn remove_all_children(b: &gtk::Box) {
    let mut child = b.first_child();
    while let Some(c) = child {
        b.remove(&c);
        child = b.first_child();
    }
}

fn load_css() {
    // Load the CSS file and add it to the provider
    let provider = CssProvider::new();
    provider.load_from_data(include_str!("style.css"));

    // Add the provider to the default screen
    gtk::style_context_add_provider_for_display(
        &Display::default().expect("Could not connect to a display."),
        &provider,
        // We use a really high priority to be able to overwrite user style sheets, as these are
        // often used for custom themes. We use the theme colors where possible, but sometimes we
        // need to color something in a way that the theme might not expect.
        1000, // gtk::STYLE_PROVIDER_PRIORITY_APPLICATION
    );
}
