use thiserror::Error;

use crate::math::{BinMathOp, MathExpr, Number, UnMathOp};

pub fn eval_math_expr(expr: MathExpr) -> Result<Number, MathEvalError> {
    match expr {
        MathExpr::Number(n) => Ok(n),
        // TODO: Non floating point arithmetic please
        MathExpr::Sym("pi") => Ok(std::f64::consts::PI),
        MathExpr::Sym("e") => Ok(std::f64::consts::E),
        MathExpr::Sym(s) => Err(MathEvalError::UnkownSymbol(s)),
        MathExpr::UnOp(UnMathOp::Minus, expr) => Ok(-eval_math_expr(*expr)?),
        MathExpr::BinOp(BinMathOp::Add, lhs, rhs) => {
            Ok(eval_math_expr(*lhs)? + eval_math_expr(*rhs)?)
        }
        MathExpr::BinOp(BinMathOp::Sub, lhs, rhs) => {
            Ok(eval_math_expr(*lhs)? - eval_math_expr(*rhs)?)
        }
        MathExpr::BinOp(BinMathOp::Mult, lhs, rhs) => {
            Ok(eval_math_expr(*lhs)? * eval_math_expr(*rhs)?)
        }
        MathExpr::BinOp(BinMathOp::Div, lhs, rhs) => {
            Ok(eval_math_expr(*lhs)? / eval_math_expr(*rhs)?)
        }
        MathExpr::BinOp(BinMathOp::Pow, lhs, rhs) => {
            Ok(eval_math_expr(*lhs)?.powf(eval_math_expr(*rhs)?))
        }
    }
}

#[derive(Debug, Error)]
pub enum MathEvalError<'a> {
    #[error("Encountered unkown symbol while evaluating: {0}")]
    UnkownSymbol(&'a str),
}

#[cfg(test)]
mod test {
    use crate::math::{
        Number,
        eval::{MathEvalError, eval_math_expr},
        parser::parse_expr_from_tokens,
        tokenizer::tokenize,
    };

    fn eval_str(input: &str) -> Result<Number, MathEvalError> {
        eval_math_expr(parse_expr_from_tokens(tokenize(input).unwrap()).unwrap())
    }

    #[test]
    fn basic_expressions() {
        assert_eq!(eval_str("1+1").unwrap(), 2.0);
        assert_eq!(eval_str("(1+1) * 6").unwrap(), 12.0);
        assert_eq!(eval_str("1/2").unwrap(), 0.5);
        assert_eq!(eval_str("2+2*3").unwrap(), 8.0);
        assert_eq!(eval_str("2+2*3**2").unwrap(), 20.0);
        assert_eq!(eval_str("pi").unwrap(), std::f64::consts::PI);
        assert_eq!(eval_str("e").unwrap(), std::f64::consts::E);
        assert_eq!(eval_str("2*pi").unwrap(), 2.0 * std::f64::consts::PI);
        assert_eq!(
            eval_str("pi * e").unwrap(),
            std::f64::consts::PI * std::f64::consts::E
        );
    }
}
