use std::iter::Peekable;

use thiserror::Error;

use crate::math::{MathToken, Number};

pub fn tokenize(input: &str) -> Result<Vec<MathToken>, MathTokenizerError> {
    let mut tokens = Vec::new();
    let mut chars = input.char_indices().peekable();
    loop {
        match chars.next() {
            Some((_, ' ')) => { /* Spaces can be ignored in the general case. */ }
            Some((_, '(')) => tokens.push(MathToken::LParen),
            Some((_, ')')) => tokens.push(MathToken::RParen),
            Some((_, '+')) => tokens.push(MathToken::Plus),
            Some((_, '-')) => tokens.push(MathToken::Minus),
            // TODO: Integer division with double //
            Some((_, '/')) => tokens.push(MathToken::Slash),
            Some((_, '*')) => match chars.peek() {
                Some((_, '*')) => {
                    chars.next();
                    tokens.push(MathToken::DoubleStar)
                }
                _ => tokens.push(MathToken::Star),
            },
            Some((_, '^')) => tokens.push(MathToken::Caret),
            Some((start_idx, c)) if c.is_alphabetic() => {
                // This is an yet unknown symbol, call the helper to find its end and return the
                // end index. Then take a slice of that symbol and store it.
                let end_idx = tokenize_sym(&mut chars, input.len());
                tokens.push(MathToken::Sym(&input[start_idx..end_idx]));
            }
            Some((start_idx, c)) if c.is_ascii_digit() || c == '.' => {
                let end_idx = tokenize_number(c == '.', &mut chars, input.len());
                let number: Number = input[start_idx..end_idx]
                    .parse()
                    .map_err(|_| MathTokenizerError::InvalidNumber(&input[start_idx..end_idx]))?;
                tokens.push(MathToken::Number(number));
            }
            Some((_, c)) => {
                return Err(MathTokenizerError::InvalidChar(c));
            }

            // No more chars, so we are done
            None => break,
        }
    }

    Ok(tokens)
}

fn tokenize_sym(
    chars: &mut Peekable<impl Iterator<Item = (usize, char)>>,
    input_end_idx: usize,
) -> usize {
    loop {
        match chars.peek() {
            // Alphabetic chars can be symbols, so continue on.
            Some((_, c)) if c.is_alphabetic() => {
                chars.next();
            }
            // Anything other then alphabetic chars means end of symbol
            Some((end_idx, _)) => return *end_idx,
            // If we reach the end of the string, that is the end of the symbol
            None => return input_end_idx,
        }
    }
}

fn tokenize_number<'a>(
    first_is_dot: bool,
    chars: &mut Peekable<impl Iterator<Item = (usize, char)>>,
    input_end_idx: usize,
) -> usize {
    let mut has_seen_dot = first_is_dot;
    loop {
        match chars.peek() {
            Some((_, '.')) if !has_seen_dot => {
                has_seen_dot = true;
                chars.next();
            }
            Some((_, c)) if c.is_ascii_digit() => {
                chars.next();
            }
            Some((end_idx, _)) => return *end_idx,
            None => return input_end_idx,
        }
    }
}

#[derive(Debug, Error)]
pub enum MathTokenizerError<'a> {
    #[error("Invalid character '{0}' in input")]
    InvalidChar(char),
    #[error("Invalid number \"{0}\" in input")]
    InvalidNumber(&'a str),
}

#[cfg(test)]
mod test {
    use crate::math::MathToken::*;
    use crate::math::tokenizer::tokenize;

    #[test]
    fn basic_examples() {
        assert_eq!(tokenize("").unwrap(), vec![]);
        assert_eq!(tokenize("1").unwrap(), vec![Number(1.)]);
        assert_eq!(tokenize("1.0").unwrap(), vec![Number(1.)]);
        assert_eq!(tokenize(".1").unwrap(), vec![Number(0.1)]);
        assert_eq!(tokenize("1+1").unwrap(), vec![Number(1.), Plus, Number(1.)]);
        assert_eq!(
            tokenize("1 + 1").unwrap(),
            vec![Number(1.), Plus, Number(1.)]
        );
        assert_eq!(
            tokenize(" 1 + 1 ").unwrap(),
            vec![Number(1.), Plus, Number(1.)]
        );

        assert_eq!(
            tokenize("(1+1)").unwrap(),
            vec![LParen, Number(1.), Plus, Number(1.), RParen]
        );
    }
}
