use std::{
    path::{Path, PathBuf},
    str::FromStr,
};

use gtk::{
    gio::{AppInfo, DesktopAppInfo, Icon, prelude::AppInfoExt},
    prelude::BoxExt,
};
use nucleo_matcher::{
    Matcher, Utf32Str,
    pattern::{AtomKind, CaseMatching, Normalization, Pattern},
};
use regex::Regex;

use crate::modules::{Module, SearchResult};

const SEARCH_THRESHOLD: u32 = 0;

pub struct DesktopResult {
    pub confidence: u32,
    pub app_info: DesktopAppInfo,
}

pub struct DesktopModule {
    pub desktop_entries: Vec<DesktopAppInfo>,
}

impl DesktopModule {
    pub fn new() -> Self {
        let desktop_entries = AppInfo::all()
            .into_iter()
            .filter_map(|entry| entry.id())
            .filter_map(|id| DesktopAppInfo::new(&id))
            .filter(|entry| {
                !entry.executable().as_os_str().is_empty()
                    && !entry.display_name().is_empty()
                    && entry.should_show()
            })
            .collect();

        Self { desktop_entries }
    }
}

impl Module for DesktopModule {
    fn results_for_search(
        &self,
        search: &str,
        matcher: &mut Matcher,
    ) -> Vec<Box<dyn SearchResult>> {
        let mut results: Vec<Box<dyn SearchResult>> = vec![];
        let pattern = Pattern::new(
            search,
            CaseMatching::Smart,
            Normalization::Smart,
            AtomKind::Fuzzy,
        );
        for app in self.desktop_entries.iter() {
            let name: String = app.name().into();
            let mut name_buf = Vec::new();
            let name = Utf32Str::new(&name, &mut name_buf);
            let score = pattern.score(name, matcher).unwrap_or(0);
            if score > SEARCH_THRESHOLD {
                results.push(Box::new(DesktopResult {
                    confidence: score,
                    app_info: app.clone(),
                }));
            }
        }

        results
    }
}

impl SearchResult for DesktopResult {
    fn confidence(&self) -> u32 {
        self.confidence
    }
    fn label(&self) -> String {
        self.app_info.name().into()
    }
    fn description(&self) -> Option<String> {
        self.app_info.description().map(Into::into)
    }

    fn icon(&self) -> Option<Icon> {
        self.app_info.icon()
    }

    fn render_detail(&self, container: &gtk::Box) {
        let text = format!("[{}] {}", self.confidence, self.app_info.name(),);
        let label = gtk::Label::builder()
            .label(text)
            .halign(gtk::Align::Start)
            .build();
        container.append(&label);
    }

    fn execute(&self) {
        let app_id = self.app_info.id();
        let mut app_exec = match self
            .app_info
            .commandline()
            .map(|p| p.to_str().map(ToOwned::to_owned))
            .flatten()
        {
            Some(app_exec) => app_exec,
            None => return,
        };

        let mut app_dir = None;
        if let Some(mut desktop_entry_path) = self.app_info.filename() {
            if let Some(desktop_entry_path) = desktop_entry_path.to_str() {
                app_exec = app_exec.replace("%k", desktop_entry_path);
            }
            desktop_entry_path.pop();
            app_dir = Some(desktop_entry_path);
        }

        app_exec = Regex::new(r"\%[uUfFdDnNickvm]")
            .unwrap()
            .replace(&app_exec, "")
            .to_string();

        let working_dir = match self
            .app_info
            .string("Path")
            .map(|path| PathBuf::from_str(&path).ok())
            .flatten()
        {
            Some(path) => Some(path),
            None => app_dir,
        };

        if self.app_info.boolean("DBusActivatable") {
            if let Some(app_id) = app_id {
                launch_cmd(vec!["gapplication", "launch", &app_id], working_dir);
            } else {
                return;
            }
        } else {
            if self.app_info.boolean("Terminal") {
                // TODO: Make terminal emulator configurable
                launch_cmd(vec!["kitty", &app_exec], working_dir);
            } else {
                if let Some(cmd) = shlex::split(&app_exec) {
                    launch_cmd(cmd, working_dir);
                }
            }
        };
    }
}

fn launch_cmd(cmd: Vec<impl AsRef<str>>, working_dir: Option<PathBuf>) {
    let cmd_prefix = vec!["systemd-run", "--user", "--scope"];
    let mut cmd_prefix = cmd_prefix.iter().map(|s| Path::new(s)).collect::<Vec<_>>();
    let cmd_suffix = cmd
        .iter()
        .map(|s| Path::new(s.as_ref()))
        .collect::<Vec<_>>();
    cmd_prefix.extend(cmd_suffix);
    let cmd = cmd_prefix;
    let env = gtk::glib::environ();
    let env = env.iter().map(|f| Path::new(f)).collect::<Vec<_>>();

    gtk::glib::spawn_async(
        working_dir,
        &cmd,
        &env,
        gtk::glib::SpawnFlags::SEARCH_PATH_FROM_ENVP | gtk::glib::SpawnFlags::SEARCH_PATH,
        None,
    )
    .unwrap();
}
